# $Id: configWatcher.py 1478 2009-06-10 09:18:20Z anton $
# $URL: https://svn.non-gnu.uvt.nl/uvt-dev/tags/sources/meresco/bzv-7.1.0/src/configWatcher.py $
from merescocore.framework import Observable

from cq2utils import DirectoryWatcher
from lxml.etree import parse

from os.path import join, isdir
from os import rename, listdir
from traceback import format_exc
import sys


class ConfigWatcherException( Exception):
    pass

class ConfigWatcher(Observable):
    """
    Config monitors a directory for files XML files being moved into it.  Each
    file moved into the directory is assumed to be in XML format.  It is read,
    parsed (using lxml) and passed on to the observers of Config using
    self.do.add(identifier=filename, lxmlNode=<parsed xml>). When the add()
    calls succeeds, the file is moved to another directory.

    Parameters

    Both the config directory and the done directory are taken as parameters from
    the constructor (__init__).  It is strongly recommended to put both
    directories on the same file system, so the move is atomic and does not
    incur data copying.

    Errors

    When an error occured, the file causing the error is also moved to the done
    directory, but an .error file is written next to it.  This file contains
    a complete stacktrace of the error.

    Notes

    The Config intentionally only listens to move events.  This avoids reading
    partial files that are still being written to.  The move operation is atomic
    and makes sure that the events of putting something into the config and
    reading it are serialized.  NOTE: move files into the config only from the
    same file system to keep its atomicity property.

    When the system starts up, the Config does not generate events for files that
    are already in the config directory.  This avoids uncontrolled bursts.
    Instead, when there are still files in the config when the system is
    restarted, either move them out and than back into the directory.
    Alternatively, one could use the method processConfigDirectory() to generate
    events for existing files programmatically.

    extract from directorywatcher.py:

    from pyinotify import WatchManager, Notifier, EventsCodes
    EVENT_MAPPING = {
        'CreateFile': EventsCodes.IN_CREATE,
        'ModifyFile': EventsCodes.IN_MODIFY,
        'MoveInFile': EventsCodes.IN_MOVED_TO,
    }
    """
    def __init__(self, reactor=None, configWatcherDirectory='', infra=None, configs=None):
        Observable.__init__(self)

        self.infra = infra
        self.logging = infra.getLoggingInstance()
        self._startLogging()

        if not isdir(configWatcherDirectory):
            msg = "directory %s does not exist" % configWatcherDirectory
            self.logger.error(msg)
            raise ConfigWatcherException(msg)

        self.configs=configs

        self._configWatcherDirectory = configWatcherDirectory
        self._watcher = DirectoryWatcher(self._configWatcherDirectory, self._processEvent, ModifyFile=True)
        self.logger.debug("configWatchterDirectory: " + configWatcherDirectory)
        reactor.addReader(self._watcher, self._watcher)

    def _startLogging(self):
        self.logger=self.logging.getLogger('meresco.configWatcher')
        self.logging.debug('Starting config logging')
        self.logger.debug('Starting config logger')

    def processConfigDirectory(self):
        for configuration in self.configs:
            processFile(configuration)

    def _processEvent(self, event):
        self.processFile(event.name)

    def processFile(self, fileName ):
        # Note: does not support trailing slashes in configWatchDirectory
        logger = self.logger
        logging = self.logging
        configs = self.configs
        infra = self.infra
        fileName = self._configWatcherDirectory + '/' + fileName

        logger.debug ("processing file: " + fileName )
        fnd = 0
        cfg = 0;

        loggingConfigFile = infra.loggingConfigfile
        if (fileName == loggingConfigFile):
            try:
                logger.info ('Updating logging configfile: "' + fileName + '"' )
                logging.config.fileConfig(fileName)

            except Exception, e:
                # logging might fail, so write to STDERR as well
                msg = 'ERROR: Could not parse loggingConfigfile: "' + fileName + '"'
                sys.stderr.write (msg + ''.join(str(e)) + "\n")
                logger.error(msg)

        else:
            # Accept (only) the config matching fileName
            for cfg in configs:
                if configs[cfg].configFile == fileName:
                    fnd = 1
                    break

            if fnd == 0:
                logger.warn('Ingoring changes in file: "' + fileName + '"')
                return

            configParser = configs[cfg].configParser

            try:
                logger.info ("Updating configfile: " + fileName )
                configParser.read(fileName)

            except Exception, e:
                logger.error('parse error file: "' + fileName + '" '+format_exc())

