# encode utf-8
from os.path import join
from merescocore.framework import Observable
from xml.sax.saxutils import escape as escapeXml
from cq2utils import autostart
from cq2utils.wrappers import wrapp
from re import compile
from lxml.etree import parse, tostring as lxmlToString, XMLSyntaxError
from StringIO import StringIO
from time import strftime, gmtime

from amara.binderytools import bind_string
from amara.bindery import is_element
from editorialdate2dateissued import createDateIssued

DOCTYPE = """<!DOCTYPE xc:content [
<!ENTITY % xhtml PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
%xhtml;
]>"""

ISO639_2b = {'EN': 'eng', 'DE': 'ger', 'FR': 'fre', 'RU': 'rus', 'ES': 'spa', 'PT': 'por'}

ROOTTAG2GENRES = {
    'article': 'info:eu-repo/semantics/article',
    'simple-article': 'info:eu-repo/semantics/article',
    'book-review': 'info:eu-repo/semantics/bookReview',
}

NSMAP = {
    "elsevier": "http://www.elsevier.com/xml/ja/dtd",
    "ce": "http://www.elsevier.com/xml/common/dtd",
}
allTagsRegexp = compile(r'<[^>]*>')
docTypeRegexp = compile(r'<![^>]*>')
xmlHeaderRegexp = compile(r'<\?[^>]*\?>')
pageSplitterRegexp = compile(r'[-+]')

def allText(amaraNode):
    # the next is the lxml (>2.0) way of getting allText
    # in 1.2 there is a bug!
    #return ''.join(lxmlNode.xpath('.//text()'))
    return allTagsRegexp.sub('', amaraNode.xml().rstrip())

def datestampNow():
    return strftime('%Y-%m-%dT%H:%M:%SZ', gmtime())

def _findTag(tagName, node, matchedTags, maxDepth):
    if node.localName and node.localName.endswith(tagName):
        matchedTags.append(wrapp(node))

    if maxDepth > 0:
        for subnode in node.childNodes:
            if is_element(subnode):
                _findTag(tagName, subnode, matchedTags, maxDepth-1)

def findTag(tagName, node, maxDepth=1):
    result = []
    _findTag(tagName, node, result, maxDepth)
    return result

def tostring(node, pretty_print=True):
    indent = 'no'
    if pretty_print:
        indent = 'yes'
    return node.xml(indent=indent)

class EditorialItemProcessor(Observable):
    def __init__(self):
        Observable.__init__(self)

    def _filenameFor(self, basedir, t3Identifier):
        if len(t3Identifier.split()) != 4:
            raise ValueError('Expected 4 parts in t3Identifier')
        t0, t1, t2, t3 = t3Identifier.split()
        return join(basedir, t1, t2, t3 + ".xml")

    def _openXmlFile(self, filename):
        xmlData = self._getStream(filename).read()
        result = StringIO()
        start = 0
        while True:
            index = xmlData.find('&', start+1)
            if index == -1:
                result.write(xmlData[start:])
                break
            result.write(xmlData[start:index])
            semicolon = xmlData.find(';', index)
            word = xmlData[index+1:semicolon]
            result.write('&')
            if word not in ['quot', 'apos', 'amp', 'lt', 'gt'] and word[0] != '#':
                result.write('amp;')
            result.write(word)

            start = semicolon

        result.seek(0)
        sanitizedXml = lxmlToString(parse(result))
        return wrapp(bind_string(sanitizedXml))

    def _getStream(self, filename):
        return open(filename)

    def _createRecordId(self, t3Identifier):
        return ':'.join(t3Identifier.split(' '))

    def _createUploadId(self, t3Identifier):
        return 'elsevier:' + self._createRecordId(t3Identifier)

    def _tagIfExists(self, tag, value):
        if not value:
            return ""
        return "<%s>%s</%s>" % (tag, escapeXml(value), tag)

    def process(self, basedir):
        editorialItem = yield
        while editorialItem != None:
            t3Identifier = editorialItem['_t3'][0]
            try:
                identifier = self._createUploadId(t3Identifier)
                article = self._openXmlFile(self._filenameFor(basedir, t3Identifier))

                #self.do.add(identifier, 'editorialItem', repr(editorialItem))
                #self.do.add(identifier, 'elsevier-article', tostring(article, pretty_print=True))
                self.do.add(identifier, 'mods', self._process(identifier, editorialItem, article))
                #self.do.add(identifier, 'meta', self._createMetaRecord(identifier, t3Identifier))
                #self.do.add(identifier, 'header', self._createHeaderRecord(t3Identifier))
                self.do.logMessage("INFO", "added %s" % (identifier,))
            except (ValueError, IOError, XMLSyntaxError), e:
                self.do.logMessage("FAILED", "%s failed with message: %s" % (identifier, str(e)))
            editorialItem = yield

    def _createMetaRecord(self, identifier, t3Identifier):
        return """<meta xmlns="http://meresco.com/namespace/harvester/meta">
  <upload><id>%(uploadId)s</id></upload>
  <record><id>%(recordId)s</id></record>
  <repository>
    <id>elsevier</id>
    <set/>
    <baseurl></baseurl>
    <repositoryGroupId>elsevier</repositoryGroupId>
    <metadataPrefix>mods</metadataPrefix>
    <collection/>
  </repository>
  </meta>""" % {
    'uploadId': escapeXml(identifier),
    'recordId': escapeXml(self._createRecordId(t3Identifier))}

    def _createHeaderRecord(self, t3Identifier):
        return """<header xmlns="http://www.openarchives.org/OAI/2.0/">
  <identifier>%(identifier)s</identifier>
  <datestamp>%(datestamp)s</datestamp>
  <setSpec>elsevier</setSpec>
</header>""" % { 'identifier': escapeXml(self._createRecordId(t3Identifier)),
                'datestamp': datestampNow()
            }

    def _process(self, identifier, editorialItem, article):
        head = self._getHead(article)
        if head == None:
            raise ValueError('Item does not contain head, identifier: ' + identifier)
        iteminfo = self._getItemInfo(article)
        if iteminfo == None:
            raise ValueError('Item does not contain item-info, identifier: ' + identifier)
        title = self._getTitles(head)
        names = self._getNames(head)
        abstract = self._getAbstract(head)
        dateIssued = self._getDateIssued(identifier, editorialItem['parent'], head)
        publisher = self._getPublisher(iteminfo)
        subject = self._getSubject(head)
        genres = self._getGenre(article)
        relatedItem = self._getRelatedItem(editorialItem)

        language = editorialItem.get('_li', [])
        if language:
            language = '<language><languageTerm authority="iso639-2b" type="code">%s</languageTerm></language>' % ISO639_2b[language[0]]

        originInfo = (publisher or dateIssued) and """<originInfo>%(publisher)s%(dateIssued)s</originInfo>""" % locals() or ""

        return """<mods xmlns:xlink="http://www.w3.org/1999/xlink" version="3.3" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns="http://www.loc.gov/mods/v3" xsi:schemaLocation="http://www.loc.gov/mods/v3 http://www.loc.gov/standards/mods/v3/mods-3-3.xsd">%(title)s%(names)s%(genres)s%(language)s%(abstract)s%(originInfo)s%(relatedItem)s%(subject)s</mods>""" % locals()


    def _getDateIssued(self, identifier, parent, head):
        date = parent.get('_dt', [None])[0]
        if date != None:
            return createDateIssued(date)
        dateAccepted = findTag('date-accepted', head)
        if len(dateAccepted) == 0:
            self.do.logMessage("WARNING", "%s Misses _dt field and has no date-accepted." % (identifier))
            return ''
        dateAccepted = dateAccepted[0]
        year = int(str(dateAccepted.year) or '0')
        month = int(str(dateAccepted.month) or '1')
        day = int(str(dateAccepted.day) or '1')
        self.do.logMessage("WARNING", "%s Misses _dt field we use date-accepted." % (identifier))
        return '<dateIssued encoding="iso8601">%04d-%02d-%02d</dateIssued>' % (year, month, day)

    def _getRelatedItem(self, editorialItem):
        titles = ''.join('<title>%s</title>' % escapeXml(journalTitle) for journalTitle in editorialItem.get('parent').get('parent').get('_jn', []))
        titleInfo = titles and '<titleInfo>%s</titleInfo>'%titles or ''
        identifier = self._getRelatedItem_identifier(editorialItem)
        part = self._getRelatedItem_part(editorialItem)        
        if titleInfo and part:
            return """<relatedItem type="host">%(titleInfo)s%(identifier)s%(part)s</relatedItem>""" % locals()
        return ''

    def _getRelatedItem_identifier(self, editorialItem):
        t1 = editorialItem['parent']['parent']['_t1'][0]
        issn = t1.split(' ')[-1]
        return '<identifier type="issn">urn:ISSN:%s-%s</identifier>' % (issn[:4], issn[4:])

    def _getRelatedItem_part(self, editorialItem):
        itemPages = editorialItem.get('_pg',[''])[0]
        pages = [p for p in pageSplitterRegexp.split(itemPages) if p]
        page = ''
        if pages:
            start = escapeXml(pages[0])
            end = escapeXml(pages[-1])
            page =  """<extent unit="page"><start>%(start)s</start><end>%(end)s</end></extent>""" % locals()
        volume = ''.join("""<detail type="volume"><number>%s</number></detail>""" % vl for vl in editorialItem.get('parent').get('_vl',[]))
        issue = ''.join("""<detail type="issue"><number>%s</number></detail>""" % iss for iss in editorialItem.get('parent').get('_is',[]))
        if volume or issue or page:
            return """<part>%(volume)s%(issue)s%(page)s</part>""" % locals()
        return ''


    def _getHead(self, parsedXml):
        """The roottag of an elsevier record differs.
        Some of the roottags we've seen so far:
            - article
            - simple-article
            - book-review
        In this method we search for the HEAD part which also
        appears with different tag-names, but they all end with
        'head' in the tag.
        """

        headTags = findTag('head', parsedXml, maxDepth=3)
        if len(headTags) == 0:
            return None

        return headTags[0]

    def _getGenre(self, parsedXml):
        rootTag = [node.localName for node in parsedXml.childNodes if is_element(node)]
        if rootTag:
            return '<genre>%s</genre>' % ROOTTAG2GENRES.get(rootTag[0], 'unknown')
        return ''

    def _getItemInfo(self, parsedXml):
        itemInfoTags = findTag('item-info', parsedXml, maxDepth=3)
        if itemInfoTags:
            return itemInfoTags[0]
        return None

    def _getPublisher(self, iteminfo):
        publishers = findTag('copyright', iteminfo)
        if len(publishers) == 0:
            return ''

        return ''.join( '<publisher>%s</publisher>' % escapeXml(p) for p in publishers if p)

    def _getNames(self, head):
        authors = findTag('author', head, maxDepth=3)

        result = ""
        for author in authors:
            innerResult = ''
            surname = author.surname

            if surname:
                innerResult += '<namePart type="family">%s</namePart>' % escapeXml(surname)
            given_name = author.given_name
            if given_name:
                innerResult += '<namePart type="given">%s</namePart>' % escapeXml(given_name)
            if innerResult:
                result += """<name type="personal">%s<role><roleTerm authority="marcrelator" type="code">aut</roleTerm></role></name>""" % innerResult
        return result

    def _getAbstract(self, head):
        paragraphs = findTag('simple-para', head, maxDepth=3)
        result = []
        for paragraph in paragraphs:
            result.append('<abstract>%s</abstract>' % allText(paragraph))
        return ''.join(result)

    def _getTitles(self, head):
        titles = findTag('title', head)

        result = []
        for title in titles:
            result.append('<title>%s</title>' % escapeXml(allText(title)))
        return '<titleInfo>%s</titleInfo>' % ''.join(result)

    def _getSubject(self, head):
        keywords = findTag('keywords', head, maxDepth=4)
        if len(keywords) == 0:
            return ''

        kw = findTag('text', keywords[0], maxDepth=3)
        if len(kw) == 0:
            return ''

        return '<subject>%s</subject>' % ''.join('<topic>%s</topic>' % escapeXml(k) for k in kw)
