from lxml.etree import parse, _ElementTree
from StringIO import StringIO
from xml.sax.saxutils import escape as escapeXml

from os.path import basename, dirname, abspath, join

from merescocore.framework import Observable

dcNamespaceMap = {
    'xsi': "http://www.w3.org/2001/XMLSchema-instance",
    'dc': 'http://purl.org/dc/elements/1.1/',
    'oai_dc': 'http://www.openarchives.org/OAI/2.0/oai_dc/'
}
modsNamespaceMap = {
    'mods': 'http://www.loc.gov/mods/v3'
}
# http://loc.gov/standards/mods/mods-dcsimple.html
class Mods2Dc(Observable):
    def __init__(self):
        Observable.__init__(self)

    def _detectAndConvert(self, anObject):
        if type(anObject) == _ElementTree:
            return self.convert(anObject)
        return anObject

    def unknown(self, method, *args, **kwargs):
        newArgs = [self._detectAndConvert(arg) for arg in args]
        newKwargs = dict((key, self._detectAndConvert(value)) for key, value in kwargs.items())
        return self.all.unknown(method, *newArgs, **newKwargs)

    def convert(self, lxmlNode):
        dc_parts = [
            self._getTitles,
            self._getCreatorsOrContributor,
            self._getSubjects,
            self._getDescriptions,
            self._getPublishers,
            self._getDates,
            self._getFormats,
            self._getIdentifiers,
            self._getLanguages,
            self._getRelations,
        ]

        result = """<oai_dc:dc xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:dc="http://purl.org/dc/elements/1.1/" xsi:schemaLocation="http://www.openarchives.org/OAI/2.0/oai_dc/ http://www.openarchives.org/OAI/2.0/oai_dc.xsd">"""
        for dc_part in dc_parts:
            result += dc_part(lxmlNode)
        result += "</oai_dc:dc>"
        try:
            return parse(StringIO(result))
        except:
            print 'Error while parsing', result
            raise

    def __str__(self):
        return 'Mods2Dc'

    def _getTitles(self, lxmlNode):
        return self._findAndBind(lxmlNode,
            '<dc:title>%s</dc:title>',
            'mods:titleInfo/mods:title/text()')

    def _getCreatorsOrContributor(self, lxmlNode):
        names = lxmlNode.xpath('mods:name', modsNamespaceMap)
        result = []
        for name in names:
            namePartsEtc = name.xpath('*[self::mods:namePart or self::mods:displayForm]/text()', modsNamespaceMap)
            if namePartsEtc:
                result.append('<dc:creator>%s</dc:creator>' % escapeXml(', '.join(namePartsEtc)))
        return '\n'.join(result)

    def _getSubjects(self, lxmlNode):
        return self._findAndBind(lxmlNode,
            '<dc:subject>%s</dc:subject>',
            'mods:subject/*[self::mods:topic or self::mods:name or self::mods:occupation]/text()',
            'mods:classification/text()')

    def _getDescriptions(self, lxmlNode):
        return self._findAndBind(lxmlNode,
            '<dc:description>%s</dc:description>',
            'mods:abstract/text()',
            'mods:note/text()',
            'mods:tableOfContents/text()')

    def _getPublishers(self, lxmlNode):
        return self._findAndBind(lxmlNode,
            '<dc:publisher>%s</dc:publisher>',
            'mods:originInfo/mods:publisher/text()')

    def _getDates(self, lxmlNode):
        return self._findAndBind(lxmlNode,
            '<dc:date>%s</dc:date>',
            'mods:originInfo/mods:dateIssued/text()',
            'mods:originInfo/mods:dateCreated/text()',
            'mods:originInfo/mods:dateCaptured/text()',
            'mods:originInfo/mods:dateOther/text()')

    def _getFormats(self, lxmlNode):
        return self._findAndBind(lxmlNode,
            '<dc:format>%s</dc:format>',
            'mods:physicalDescription/mods:internetMediaType/text()',
            'mods:physicalDescription/mods:extent/text()',
            'mods:physicalDescription/mods:form/text()')

    def _getIdentifiers(self, lxmlNode):
        return self._findAndBind(lxmlNode,
            '<dc:identifier>%s</dc:identifier>',
            'mods:identifier/text()',
            'mods:location/mods:url/text()')

    def _getLanguages(self, lxmlNode):
        return self._findAndBind(lxmlNode,
            '<dc:language>%s</dc:language>',
            'mods:language/mods:languageTerm/text()')

    def _getRelations(self, lxmlNode):
        return self._findAndBind(lxmlNode,
            '<dc:relation>%s</dc:relation>',
            'mods:relatedItem/mods:titleInfo/mods:title/text()')

    def _findAndBind(self, node, template, *xpaths):
        items = []
        for p in xpaths:
            items += node.xpath(p, modsNamespaceMap)
        return '\n'.join(template % escapeXml(item) for item in items)
