from merescocore.framework import Observable

from storage.storage import unescapeName
from xml.sax.saxutils import escape as xml_escape
from os.path import basename
from StringIO import StringIO
from lxml.etree import parse, tostring

EXTENSION = '.record'

class XmlDocumentAdapter(Observable):
    """
    This class wraps a single lxmlNode into a document as expected by many of
    the Meresco components.  It splits the identifier into a repository and a
    record identifier by splitting on the first collon. Also, it only responds
    to identifiers ending with '.record'. As an example:

        <repository id>:<record id>.record.

    will be wrapped in a <document> tag, with a <header> and a <meta> containing
    the repository and record id extracted from the input identifier.  The
    original record is put in the <metadata> tag.

    Adaptation for Tilburg University:
    The filename is expected to be structured as follows:

        <collection id>:<repository group id>:<repository id>:<record id>.record

    A <header> element is not included in the document. The four id's are extracted from
    the filename and used for generating a <meta> part.
    """

    def _splitFilename(self, aString):
        if aString.endswith(EXTENSION):
            parts = aString[:-len(EXTENSION)].split(':', 3)
            if len(parts) == 4:
                collection, repositoryGroupId, repository, identifier = parts
                return collection, repositoryGroupId, repository, unescapeName(identifier)
        return ('', '')

    def _createRecord(self, collection, repositoryGroupId, repository, identifier, lxmlNode):
#        header = xml_escape("""<oai:header xmlns:oai="http://www.openarchives.org/OAI/2.0/">
#            <oai:identifier>%(identifier)s</oai:identifier>
#        </oai:header>""" % locals())

        meta = xml_escape("""<meta:meta xmlns:meta="http://meresco.com/namespace/harvester/meta">
  <meta:upload><meta:id>%(repository)s:%(identifier)s</meta:id></meta:upload>
  <meta:record>
    <meta:id>%(identifier)s</meta:id>
  </meta:record>
  <meta:repository>
    <meta:id>%(repository)s</meta:id>
    <meta:repositoryGroupId>%(repositoryGroupId)s</meta:repositoryGroupId>
    <meta:collection>%(collection)s</meta:collection>
  </meta:repository>
</meta:meta>""" % locals())
        metadata = xml_escape("""<oai:metadata xmlns:oai="http://www.openarchives.org/OAI/2.0/">%s</oai:metadata>""" % tostring(lxmlNode))

        return parse(StringIO("""<document xmlns="http://meresco.com/namespace/harvester/document"><part name="meta">%(meta)s</part><part name="metadata">%(metadata)s</part></document>""" % locals()))
#        return parse(StringIO("""<document xmlns="http://meresco.com/namespace/harvester/document"><part name="header">%(header)s</part><part name="meta">%(meta)s</part><part name="metadata">%(metadata)s</part></document>""" % locals()))

    def add(self, identifier=None, lxmlNode=None):
        collection, repositoryGroupId, repositoryId, recordId = self._splitFilename(identifier)
        if collection == '' or repositoryGroupId == '' or repositoryId == '' or recordId == '':
            return

        self.do.add(identifier=repositoryId+':'+recordId,
            lxmlNode=self._createRecord(collection, repositoryGroupId, repositoryId, recordId, lxmlNode))
