#coding: utf-8
from cq2utils import CQ2TestCase, CallTrace
from os.path import abspath, dirname, join
from weightless import compose
from StringIO import StringIO
#from lxml.etree import parse, tostring, XMLParser, XMLSchema
from lxml.etree import tostring, XMLParser, XMLSchema, parse as lxmlParse
from amara.binderytools import bind_stream
from cq2utils.wrappers import wrapp
from difflib import unified_diff
from time import strftime, gmtime

from bzv import EditorialItemProcessor

mydir = dirname(abspath(__file__))
modsschema = XMLSchema(lxmlParse(open('../schemas/mods-3-3.xsd')))

def parse(stream):
    return wrapp(bind_stream(stream))

def parseString(data):
    return parse(StringIO(data))

def prettyPrint(data):
    return parseString(data).xml(indent='yes')

class EditorialItemProcessorTest(CQ2TestCase):

    def assertPresentWS(self, string, substring, *args, **kwargs):
        def _strip(aString):
            result = aString
            for i in [' ', '\t', '\n']:
                result = result.replace(i, '')
            return result
        self.assertTrue(_strip(substring) in _strip(string), *args, **kwargs)

    def testOne(self):
        observer = CallTrace('Observer')
        processor = EditorialItemProcessor()
        processor._getStream = lambda x: StringIO(TESTDATA_XML)
        processor.addObserver(observer)
        generator = compose(processor.process(join(mydir, 'data')))
        generator.next()
        generator.send(TESTDATA)
        generator.close()
        self.assertEquals(2, len(observer.calledMethods))
        self.assertEquals(['add', 'logMessage'], [m.name for m in observer.calledMethods])

        method = observer.calledMethods[0]
        self.assertEquals(3, len(method.args))

        partname = method.args[1]
        self.assertEquals('mods', partname)
        mods = method.args[2]

        self.assertPresentWS(mods, """<titleInfo><title>Intergenerational discounting: A case from Hong Kong</title></titleInfo>""", mods)
        self.assertPresentWS(mods, """<language><languageTerm authority="iso639-2b" type="code">eng</languageTerm></language>""")
        self.assertPresentWS(mods, """<abstract>In debating sustainable development""", mods)

        self.assertPresentWS(mods, """<subject><topic>Sustainable development</topic><topic>Discount rate</topic><topic>Land value</topic><topic>Land tenure</topic></subject>""",mods)

    def testDateIssuedAllways_dt(self):
        processor = EditorialItemProcessor()
        inputXml = parse(StringIO(TESTDATA_XML))
        self.assertEqualsWS('<dateIssued qualifier="inferred" encoding="iso8601">2008-11-04</dateIssued>',
            processor._getDateIssued('identifier', {'_dt':['20081104']}, processor._getHead(inputXml)))

    def testDateIssuedFallback(self):
        processor = EditorialItemProcessor()
        inputXml = parse(StringIO(TESTDATA_XML_RAW % ''))
        self.assertEqualsWS('<dateIssued qualifier="inferred" encoding="iso8601">2006-03</dateIssued>',
            processor._getDateIssued('identifier', {'_dt': ['200603']},
                processor._getHead(inputXml)))

    def testDateIssuedFallbackToHead(self):
        processor = EditorialItemProcessor()
        inputXml = parse(StringIO(TESTDATA_XML))
        self.assertEqualsWS('<dateIssued encoding="iso8601">2008-03-20</dateIssued>',
            processor._getDateIssued('identifier', {},
                processor._getHead(inputXml)))

    def testNoDateIssued(self):
        processor = EditorialItemProcessor()
        inputXml = parse(StringIO(TESTDATA_XML_RAW % ''))
        self.assertEqualsWS('',
            processor._getDateIssued('identifier', {},
                processor._getHead(inputXml)))

    def testGetNames(self):
        processor = EditorialItemProcessor()
        self.assertEqualsWS('<name type="personal"><namePart type="family">SURNAME</namePart><namePart type="given">GIVEN-NAME</namePart><role><roleTerm authority="marcrelator" type="code">aut</roleTerm></role></name><name type="personal"><namePart type="family">surname</namePart><role><roleTerm authority="marcrelator" type="code">aut</roleTerm></role></name>', processor._getNames(processor._getHead(parse(StringIO(TESTDATA_XML)))))

    def testFilenameFor(self):
        processor = EditorialItemProcessor()
        self.assertEquals('PATH/B/C/D.xml', processor._filenameFor('PATH', 'A B C D'))

    def testEscapeUnknownEntities(self):
        processor = EditorialItemProcessor()
        processor._getStream = lambda x: StringIO('<tag>&beta;&lt;&gt;</tag>')

        parsedXml = processor._openXmlFile('Name is ignored in this test')
        self.assertEquals('<?xml version="1.0" encoding="UTF-8"?>\n<tag>&amp;beta;&lt;&gt;</tag>', parsedXml.xml())

    def testFiles(self):
        self.assertMods("elsevier:CXT0242A:00092614:04570004:08004910")
        self.assertMods("elsevier:CXT0242A:01973975:00320003:08000027")
        self.assertMods("elsevier:CXT0242A:01973975:00320003:08000301")
        self.assertMods("elsevier:CXT0242A:0926860X:03420001:08001646")
        self.assertMods("elsevier:CXT0242A:09409602:01900003:08000289")
        self.assertMods("elsevier:CXT0242A:1464391X:00100003:08700618")
        self.assertMods("elsevier:CXT0239A:01964399:unassign:08000226")

    def assertMods(self, id):
        editorialItem = eval(open('data/%s.editorialItem' % id).read())
        processor = EditorialItemProcessor()
        processor._getStream = lambda x: open('data/%s.original.xml' % id)
        original = processor._openXmlFile('ignored for tests')

        result = prettyPrint(processor._process(id, editorialItem, original))
        expectedMods = prettyPrint(open('data/%s.mods.xml' % id).read())
        modsschema.validate(lxmlParse(StringIO(result)))

        if modsschema.error_log.last_error:
            for nr, line in enumerate(result.split('\n')):
                print nr+1, line
            self.fail(modsschema.error_log.last_error)
        self.assertEqualsWithDiff(expectedMods, result)

    def assertEqualsWithDiff(self, expected, result):
        diffs = list(unified_diff(
            [x.strip() for x in result.split('\n') if x.strip()],
            [x.strip() for x in expected.split('\n') if x.strip()],
            fromfile='result', tofile='expected', lineterm=''))
        self.assertFalse(diffs, '\n' + '\n'.join(diffs))

    def testRecordId(self):
        processor = EditorialItemProcessor()
        self.assertEquals('CXT0242A:01973975:00320003:08000027', processor._createRecordId('CXT0242A 01973975 00320003 08000027'))
        self.assertEquals('elsevier:CXT0242A:01973975:00320003:08000027', processor._createUploadId('CXT0242A 01973975 00320003 08000027'))

    def testValidISSN(self):
        #just for fun
        issn = '1464391X'
        checksum = 11- (sum(map(lambda (x,y):x*y, zip([int(i) for i in issn[:-1]], range(8,1,-1))))% 11)
        checksum = checksum == 10 and 'X' or str(checksum)
        self.assertEquals(checksum, issn[-1])

    def testErrorWhileRetrievingXml(self):
        def raiseIOError(name):
            raise IOError('Name "%s" does not exist.' % name)
        observer = CallTrace('observer')
        processor = EditorialItemProcessor()
        processor.addObserver(observer)
        processor._getStream = raiseIOError
        generator = compose(processor.process(join(mydir, 'data')))
        generator.next()
        generator.send(TESTDATA)
        generator.close()
        self.assertEquals(1, len(observer.calledMethods))
        self.assertEquals('logMessage(\'FAILED\', \'elsevier:CXT0242A:01973975:00320003:08000027 failed with message: Name "%s/data/01973975/00320003/08000027.xml" does not exist.\')' % mydir, str(observer.calledMethods[0]))

    def testErrorWhileCreatingDateIssued(self):
        def raiseValueError(*args):
            raise ValueError('BAD')
        observer = CallTrace('observer')
        processor = EditorialItemProcessor()
        processor.addObserver(observer)
        processor._getStream = lambda x: StringIO(TESTDATA_XML_RAW)
        processor._getDateIssued = raiseValueError
        generator = compose(processor.process(join(mydir, 'data')))
        generator.next()
        generator.send(TESTDATA)
        generator.close()
        self.assertEquals(1, len(observer.calledMethods))
        self.assertEquals('logMessage(\'FAILED\', \'elsevier:CXT0242A:01973975:00320003:08000027 failed with message: BAD\')', str(observer.calledMethods[0]))



TESTDATA = {
    '_t3': ['CXT0242A 01973975 00320003 08000027'],
    '_ti': ['IGNORED'],
    '_au': ['SURNAME, GIVEN-NAME', 'surname, given-name'],
    '_li': ['EN'],
    '_ab': ['LINE 1. LINE 2.'],
    '_kw': ['keyword_0', 'keyword_1', 'keyword_2'],
    'parent': {
        '_t2': ['CXT0242A 01973975 00320003'],
        '_dt': ['20081104'],
        'parent': {
            '_t1': ['CXT0242A 01973975'],
            'parent': {
                '_t0': ['CXT0242A'],
            }
        }
    }
}
TESTDATA_DT_INVALID = {
    '_t3': ['CXT0242A 01973975 00320003 08000027'],
    '_ti': ['IGNORED'],
    '_au': ['SURNAME, GIVEN-NAME', 'surname, given-name'],
    '_li': ['EN'],
    '_ab': ['LINE 1. LINE 2.'],
    '_kw': ['keyword_0', 'keyword_1', 'keyword_2'],
    'parent': {
        '_t2': ['CXT0242A 01973975 00320003'],
        '_dt': ['20081104/20071212'],
        'parent': {
            '_t1': ['CXT0242A 01973975'],
            'parent': {
                '_t0': ['CXT0242A'],
            }
        }
    }
}
TESTDATA_XML_RAW = """<?xml version = '1.0' encoding = 'utf-8'?>
<!DOCTYPE article PUBLIC "-//ES//DTD journal article DTD version 5.0.1//EN//XML" "art501.dtd">
<article xmlns="http://www.elsevier.com/xml/ja/dtd" xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:ce="http://www.elsevier.com/xml/common/dtd" xmlns:sb="http://www.elsevier.com/xml/common/struct-bib/dtd" docsubtype="fla" xml:lang="en" version="5.0">
   <item-info>
      <jid>HAB</jid>
      <aid>731</aid>
      <ce:pii>S0197-3975(07)00057-4</ce:pii>
      <ce:doi>10.1016/j.habitatint.2007.08.015</ce:doi>
      <ce:copyright type="full-transfer" year="2007">Elsevier Ltd</ce:copyright>
   </item-info>
   <head>
      <ce:title>Intergenerational discounting: A case from Hong Kong</ce:title>
      <ce:author-group>
         <ce:author>
            <ce:given-name>GIVEN-NAME</ce:given-name>
            <ce:surname>SURNAME</ce:surname>
            <ce:cross-ref refid="aff1">
               <ce:sup loc="post">a</ce:sup>
            </ce:cross-ref>
            <ce:cross-ref refid="cor1">
               <ce:sup loc="post">⁎</ce:sup>
            </ce:cross-ref>
            <ce:e-address type="email">skwongb@hku.hk</ce:e-address>
         </ce:author>
         <ce:author>
            <ce:surname>surname</ce:surname>
            <ce:cross-ref refid="aff1">
               <ce:sup loc="post">a</ce:sup>
            </ce:cross-ref>
         </ce:author>
         <ce:affiliation id="aff1">
            <ce:label>a</ce:label>
            <ce:textfn>Department of Real Estate and Construction, The University of Hong Kong, Hong Kong</ce:textfn>
         </ce:affiliation>
         <ce:affiliation id="aff2">
            <ce:label>b</ce:label>
            <ce:textfn>The Bartlett School of Graduate Studies, University College London, UK</ce:textfn>
         </ce:affiliation>
         <ce:correspondence id="cor1">
            <ce:label>⁎</ce:label>
            <ce:text>Corresponding author. Fax: +852<ce:hsp sp="0.16"/>2559<ce:hsp sp="0.16"/>9457.</ce:text>
         </ce:correspondence>
      </ce:author-group>
      <ce:date-received day="13" month="12" year="2007"/>
      <ce:date-revised day="15" month="3" year="2008"/>
      %s
      <ce:abstract class="author">
         <ce:section-title>Abstract</ce:section-title>
         <ce:abstract-sec>
            <ce:simple-para view="all">In debating sustainable development issues such as climate changes and nuclear waste disposal, policymakers usually base their decisions on cost-and-benefit analysis (CBA), which evaluates tradeoffs between costs and benefits spanning over a century. Fundamental to this approach is the choice of an intergenerational discount rate for a society. Observable market rates, however, are mostly available up to 30 years only (e.g. US bonds), which do not necessarily reflect the long-term discount rate for a society to deal with sustainable development. A number of studies have demonstrated that a small change in the discount rate can drastically alter the CBA's outcome. This paper seeks to unveil the market's intergenerational discount rate by analyzing a very “long-lived” and inheritable asset—land property. By examining a mix of lease tenures (50, 75, 99, and 999 years) and their transactions, we found that 999-year property commands a significant premium of 5.74 over 99-year property from 1992 through 2006. The premium implies an intergenerational discount rate of 4.31 p.a. on average, which is significantly lower than the intragenerational discount rate. The discount rate so unveiled would contribute to a more informed intergenerational decision analysis, such as evaluating the cost-effectiveness of environmental legislation and assessing very long-term pollution damages in court. The result also has important implications for real estate valuation practices in China, as all land in Mainland China and Hong Kong is leasehold.</ce:simple-para>
         </ce:abstract-sec>
      </ce:abstract>
      <ce:keywords class="keyword">
         <ce:section-title>Keywords</ce:section-title>
         <ce:keyword>
            <ce:text>Sustainable development</ce:text>
         </ce:keyword>
         <ce:keyword>
            <ce:text>Discount rate</ce:text>
         </ce:keyword>
         <ce:keyword>
            <ce:text>Land value</ce:text>
         </ce:keyword>
         <ce:keyword>
            <ce:text>Land tenure</ce:text>
         </ce:keyword>
      </ce:keywords>
   </head>
</article>"""

TESTDATA_XML = TESTDATA_XML_RAW % '<ce:date-accepted day="20" month="3" year="2008"/>'

